-----------------------------------------------------------------------------------------------
-- ColorUtils Module Definition
-- Provides utilties for functions that don't need to exist in IronColorPicker.
-----------------------------------------------------------------------------------------------

local ColorUtils = {}
local IronColorPicker = Apollo.GetAddon("IronColorPicker")

---------------------------------------------------------------------------------------------------
function ColorUtils:Init(parent)
---------------------------------------------------------------------------------------------------
	Apollo:LinkAddon(parent, self)
end	

---------------------------------------------------------------------------------------------------
function ColorUtils:Clamp(nValue, nMax, nMin)
---------------------------------------------------------------------------------------------------
	return (nValue > nMax) and nMax or (nValue < nMin) and nMin or nValue
end

-----------------------------------------------------------------------------------------------
-- Expects: r, g, b in range 0 .. 1.0
-- Returns: table with h, s, v
-----------------------------------------------------------------------------------------------
function ColorUtils:RgbToHsv(r, g, b)
-----------------------------------------------------------------------------------------------	
	local mincolor, maxcolor = math.min(r, g, b), math.max(r, g, b)
	local ch, cs, cv = 0, 0, maxcolor
	if maxcolor > 0 then -- technically ch is undefined if cs is zero
		local delta = maxcolor - mincolor
		cs = delta / maxcolor
		if delta > 0 then -- don't allow divide by zero
			if r == maxcolor then
				ch = (g - b) / delta -- between yellow and magenta
			elseif g == maxcolor then
				ch = 2 + ((b - r) / delta) -- between cyan and yellow
			else
				ch = 4 + ((r - g) / delta) -- between magenta and cyan
			end
		end
		if ch < 0 then ch = ch + 6 end -- correct for negative values
		ch = ch / 6 -- and finally adjust range 0 to 1.0
	end
	return { H = ch, S = cs, V = cv }
end

-----------------------------------------------------------------------------------------------
-- Convert h, s, l input values into r, g, b return values
-- All values are in the range 0 to 1.0
-----------------------------------------------------------------------------------------------
function ColorUtils:HsvToRgb(ch, cs, cv)
-----------------------------------------------------------------------------------------------
	local r, g, b = cv, cv, cv
	if cs > 0 then -- if cs is zero then grey is returned
		local h = ch * 6; local sextant = math.floor(h) -- figure out which sextant of the color wheel
		local fract = h - sextant -- fractional offset into the sextant
		local p, q, t = cv * (1 - cs), cv * (1 - (cs * fract)), cv * (1 - (cs * (1 - fract)))
		if sextant == 0 then
			r, g, b = cv, t, p
		elseif sextant == 1 then
			r, g, b = q, cv, p
		elseif sextant == 2 then
			r, g, b = p, cv, t
		elseif sextant == 3 then
			r, g, b = p, q, cv
		elseif sextant == 4 then
			r, g, b = t, p, cv
		else
			r, g, b = cv, p, q
		end
	end
	return {R = r, G = g, B = b}
end

-----------------------------------------------------------------------------------------------
function ColorUtils:RgbToString(r, g, b)
-----------------------------------------------------------------------------------------------
	return string.format("%02x%02x%02x", math.floor(r * 255 + 0.5), math.floor(g * 255 + 0.5), math.floor(b * 255 + 0.5))
end

-----------------------------------------------------------------------------------------------
function ColorUtils:HsvToString(h, s, v)
-----------------------------------------------------------------------------------------------
	local tColors = self:HsvToRgb(h, s, v)
	return "ff"..self:RgbToString(tColors.R, tColors.G, tColors.B)
end

-----------------------------------------------------------------------------------------------
function ColorUtils:ValueToRgb(strValue)
-----------------------------------------------------------------------------------------------
	return 
	{ 
	  R = tonumber(strValue:sub(1,2), 16), 
	  G = tonumber(strValue:sub(3,4), 16), 
	  B = tonumber(strValue:sub(5,6), 16)
	} 
end

-----------------------------------------------------------------------------------------------
function ColorUtils:ConcatTables(tLhs, tRhs)
-----------------------------------------------------------------------------------------------
	for k,v in pairs(tRhs) do
	    tLhs[k] = v
	end
	return tLhs
end	

IronColorPicker.ColorUtils = ColorUtils