-----------------------------------------------------------------------------------------------
-- Client Lua Script for IronColorPicker
-----------------------------------------------------------------------------------------------
require "Window"
 
-----------------------------------------------------------------------------------------------
-- IronColorPicker Module Definition
-----------------------------------------------------------------------------------------------
local IronColorPicker = {}
 
-----------------------------------------------------------------------------------------------
-- Constants
-----------------------------------------------------------------------------------------------
local k_DefaultColor = CColor.new(1, 0, 0, 1)
local k_Hue          = "Hue"
local k_Preview      = "Preview"
local k_Dial         = "Dial"
local k_Selector     = "Selector"
local k_Gradient     = "Gradient"
local k_Overlay      = "Overlay"
local k_Main         = "Main"

-- Values
local k_H = "H"
local k_V = "V"
local k_S = "S"

local k_R = "R"
local k_G = "G"
local k_B = "B"

local k_Hex = "HEX"

local k_LeftMouse    = 0
local k_RightMouse   = 1

-----------------------------------------------------------------------------------------------
-- Initialization
-----------------------------------------------------------------------------------------------
function IronColorPicker:new(o)
-----------------------------------------------------------------------------------------------
    o = o or {}
    setmetatable(o, self)
    self.__index = self 
    
    self.ColorUtils = nil

    -- initialize variables here
    return o
end

-----------------------------------------------------------------------------------------------
function IronColorPicker:Init()
-----------------------------------------------------------------------------------------------
	local bHasConfigureFunction = false
	local strConfigureButtonText = ""
	local tDependencies = {
	}
    Apollo.RegisterAddon(self, bHasConfigureFunction, strConfigureButtonText, tDependencies)
end

-----------------------------------------------------------------------------------------------
function IronColorPicker:OnIronColorPickerOn()
-----------------------------------------------------------------------------------------------
	self.wndMain:Invoke() -- show the window
end

-----------------------------------------------------------------------------------------------
function IronColorPicker:OnClose( wndHandler, wndControl, eMouseButton )
-----------------------------------------------------------------------------------------------
	self.wndMain:Close()
end

-----------------------------------------------------------------------------------------------
-- Callback for a color being selected. The user can do whatever they want with the
-- resulting table, that includes R, G, B, H, S, V and a formatted string.
-----------------------------------------------------------------------------------------------
function IronColorPicker:OnSelect( wndHandler, wndControl, eMouseButton )
-----------------------------------------------------------------------------------------------
	if self.callback ~= nil and type(self.callback) == "function" then
		self.wndMain:Close()

		local tColors = {}

		-- Dump raw values
		for i,v in ipairs({k_R, k_G, k_B, k_H, k_S, k_V}) do
			tColors[v] = self:GetColor(v) * self.tValues[v].Max
		end

		-- Dump formatted hex string
		tColors[k_Hex] = self.ColorUtils:RgbToString(self:GetColor(k_R), self:GetColor(k_G), self:GetColor(k_B))

		self.callback(tColors)
	end
end

-----------------------------------------------------------------------------------------------
function IronColorPicker:OnLoad()
-----------------------------------------------------------------------------------------------
    -- load our form file
	self.xmlDoc = XmlDoc.CreateFromFile("IronColorPicker.xml")
	self.xmlDoc:RegisterCallback("OnDocLoaded", self)
	Apollo.LoadSprites("IronColorPickerSprites.xml", "CP_Sprites")
end

-----------------------------------------------------------------------------------------------
function IronColorPicker:OnDocLoaded()
-----------------------------------------------------------------------------------------------
	if self.xmlDoc ~= nil and self.xmlDoc:IsLoaded() then
	    self.wndMain = Apollo.LoadForm(self.xmlDoc, "Main", nil, self)
		if self.wndMain == nil then
			Apollo.AddAddonErrorText(self, "Could not load the main window for some reason.")
			return
		end

		Apollo.RegisterSlashCommand("icp", "OnIronColorPickerOn", self)

		-- Create function table for UI updates
		self.UpdateTable = {}
	    self.UpdateTable[k_Hue] 	= function(...) return self:UpdateHue     (...) end
	    self.UpdateTable[k_Overlay] = function(...) return self:UpdateOverlay (...) end
	    self.UpdateTable[k_Main] 	= function(...) return self:UpdateMain    (...) end

		self.wndMain:Show(false)
		self.ColorUtils:Init(self)

		self.tCoords = {}
		self.tValues = {}

		-- Setup all UI elements that need a sprite or something other than just default values
		self:LoadCoord(k_Hue, 		{strDefaultSprite = "CP_Sprites:colors"  })
		self:LoadCoord(k_Gradient, 	{strDefaultSprite = "CP_Sprites:gradient"})
		self:LoadCoord(k_Overlay, 	{strDefaultSprite = "CP_Sprites:overlay",  nDefaultBGColor = k_DefaultColor})
		self:LoadCoord(k_Preview,   {nDefaultBGColor = k_DefaultColor})
		self:LoadCoord(k_Selector)
		self:LoadCoord(k_Dial, 		{nDefaultRotation = 90})
		self:LoadCoord(k_Main)

		-- Setup the text boxes for HSV
		self:LoadValues("HSV", 
		{
			{ name = k_H, max = 360, min = 0, value = 0   },
			{ name = k_S, max = 100, min = 0, value = 100 },
			{ name = k_V, max = 100, min = 0, value = 100 },
		})

		-- Setup the text boxes for RGB
		self:LoadValues("RGB", 
		{
			{ name = k_R, max = 255, min = 0, value = 255 },
			{ name = k_G, max = 255, min = 0, value = 0   },
			{ name = k_B, max = 255, min = 0, value = 0   },
		})

		-- Setup the initial value for hex
		self:LoadHex("HEX", "ff0000")

		-- Update the UI with initial values
		self:UpdateValues("HSV", true)
	end
end

-----------------------------------------------------------------------------------------------
-- Consumer Facing API
-----------------------------------------------------------------------------------------------
function IronColorPicker:ShowColor(strColor, funcCallback)
-----------------------------------------------------------------------------------------------
	self.callback = funcCallback
	local tColors = self.ColorUtils:ValueToRgb(strColor)
	self:SetColorValue(k_R, tColors.R)
	self:SetColorValue(k_G, tColors.G)
	self:SetColorValue(k_B, tColors.B)
	self:UpdateValues("RGB", true)
	self.wndMain:Show(true)
	self.wndMain:ToFront()
end

-----------------------------------------------------------------------------------------------
-- UI Callbacks
-- UpdateHue
-- UpdateOverlay
-- UpdateMain
-- SetProperty
-- SetColorValue
-- UpdateValues
-- OnValueEdited
-- OnValueChanged
-----------------------------------------------------------------------------------------------

-----------------------------------------------------------------------------------------------
-- Update for mouse down and inside hue.
-----------------------------------------------------------------------------------------------
-- x           : x position
-- y 		   : y position
-- bArticifial : is this update artificial? (the result of updating a text box)
-----------------------------------------------------------------------------------------------
function IronColorPicker:UpdateHue(x, y, bArtificial)
-----------------------------------------------------------------------------------------------
	-- Update values
	if bArtificial == false then
		local hue = self.ColorUtils:Clamp(self:GetY(k_Hue) / self:GetHeight(k_Hue), 1, 0)
		hue = (hue == 1) and 0 or hue
		self:SetColor(k_H, hue) 
		self:UpdateValues("HSV", false)
	end

	-- Update dial position
	local left, top, right, bottom = self.tCoords[k_Dial].Widget:GetAnchorOffsets()
	top = self:GetY(k_Hue) - self:GetHeight(k_Dial)/2
	self.tCoords[k_Dial].Widget:SetAnchorOffsets(left, top, right, top + self:GetHeight(k_Dial))

	-- Update overlay and preview color
	self:GetWidget(k_Overlay):SetBGColor(self.ColorUtils:HsvToString(self:GetColor(k_H), 1, 1))
	self:GetWidget(k_Preview):SetBGColor(CColor.new(self:GetColor(k_R), self:GetColor(k_G), self:GetColor(k_B), 1)) 
end

-----------------------------------------------------------------------------------------------
-- Update for mouse down and inside overlay.
-----------------------------------------------------------------------------------------------
-- x           : x position
-- y 		   : y position
-- bArticifial : is this update artificial? (the result of updating a text box)
-----------------------------------------------------------------------------------------------
function IronColorPicker:UpdateOverlay(x, y, bArtificial)
-----------------------------------------------------------------------------------------------
	if bArtificial == false then 
		self:SetColor(k_S, self.ColorUtils:Clamp(self:GetX(k_Overlay) / self:GetWidth(k_Overlay),  1, 0))
		self:SetColor(k_V, self.ColorUtils:Clamp((self:GetHeight(k_Overlay) - self:GetY(k_Overlay)) / self:GetHeight(k_Overlay), 1, 0))
		self:UpdateValues("HSV", false)
	end 

	-- Update selector position
	local left, top, right, bottom = self:GetWidget(k_Selector):GetAnchorOffsets()
	top  = self:GetY(k_Overlay) - self:GetHeight(k_Selector)/3
	left = self:GetX(k_Overlay) - self:GetWidth(k_Selector)/3
	self:GetWidget(k_Selector):SetAnchorOffsets(left, top, left + self:GetWidth(k_Selector), top + self:GetHeight(k_Selector))

	-- Update preview color
	self:GetWidget(k_Preview):SetBGColor(CColor.new(self:GetColor(k_R), self:GetColor(k_G), self:GetColor(k_B), 1))
end

-----------------------------------------------------------------------------------------------
-- Update function for mouse down in a UI element that is not hue or overlay.
-----------------------------------------------------------------------------------------------
-- x           : x position
-- y 		   : y position
-- bArticifial : is this update artificial? (the result of updating a text box)
-----------------------------------------------------------------------------------------------
function IronColorPicker:UpdateMain(x, y, bArtificial)
-----------------------------------------------------------------------------------------------
	-- Hue is being dragged, but we're outside hue
	if self:GetMouse(k_Hue) and self:GetWidget(k_Hue):ContainsMouse() == false then
		local l, t, r, b = self:GetWidget(k_Hue):GetParent():GetRect() 
		self:SetProperty(k_Hue, x, y - t, false)

	-- Overlay is being dragged, but we're outside overlay
	elseif self:GetMouse(k_Overlay) and self:GetWidget(k_Overlay):ContainsMouse() == false then
		local l, t, r, b = self:GetWidget(k_Overlay):GetParent():GetRect() 
		self:SetProperty(k_Overlay, x - l, y - t, false)
	end
end

-----------------------------------------------------------------------------------------------
-- Sets the bounded x and y positions for a given ui element, and performs
-- the update for the given element.
-----------------------------------------------------------------------------------------------
-- strName     : the name of the widget whose properties are being set
-- x           : x position
-- y 		   : y position
-- bArticifial : is this update artificial? (the result of updating a text box)
-----------------------------------------------------------------------------------------------
function IronColorPicker:SetProperty(strName, x, y, bArtificial)
-----------------------------------------------------------------------------------------------
	-- Clamp values to be correctly inside bounds (needed for main updates outside hue/overlay)
	self:SetX(strName, self.ColorUtils:Clamp(x, self:GetWidth (strName), 0))
	self:SetY(strName, self.ColorUtils:Clamp(y, self:GetHeight(strName), 0))

	-- Index into the function table to do the rest of the work
	self.UpdateTable[strName](x, y, bArtificial)
end

-----------------------------------------------------------------------------------------------
-- Set one of HSV or RGB, doing bounds checking and updating both the UI
-- and the table that stores the values in the background.
-----------------------------------------------------------------------------------------------
-- strName : name of the color value, one of [ H, S, V, R, G, B ]
-- nValue  : the value being set, not in range 0 .. 1.0 but instead 0 .. max
-----------------------------------------------------------------------------------------------
function IronColorPicker:SetColorValue(strName, nValue)
----------------------------------------------------------------------------------------------- 
	local max = self.tValues[strName].Max
	local min = self.tValues[strName].Min
	local wid = self.tValues[strName].Widget
	if nValue <= max and nValue >= min then
		wid:SetText(string.format("%d", nValue))
		wid:SetSel(wid:GetMaxTextLength()) -- gets rid of weird behavior when SetText is called
		self:SetColor(strName, nValue/max)
	end
end

-----------------------------------------------------------------------------------------------
-- Recalculates either all of HSV or RGB based on the value that was updated.
-----------------------------------------------------------------------------------------------
-- strKey 		 : key that tells us whether is update is HSV or RGB
-- bUpdateCoords : should the hue and overlay be updated as well? (removes double updates)
-----------------------------------------------------------------------------------------------
function IronColorPicker:UpdateValues(strKey, bUpdateCoords)
----------------------------------------------------------------------------------------------- 
	local tColors = {}

	-- Only update either RGB or HSV, not both
	if string.match(strKey, "[HSV]") then
		tColors = self.ColorUtils:HsvToRgb(self:GetColor(k_H), self:GetColor(k_S), self:GetColor(k_V))
		self.ColorUtils:ConcatTables(tColors, self:GetHSVTable())
	elseif string.find(strKey, "[RGB]") then
		tColors = self.ColorUtils:RgbToHsv(self:GetColor(k_R), self:GetColor(k_G), self:GetColor(k_B))
		self.ColorUtils:ConcatTables(tColors, self:GetRGBTable())
	end

	-- Set the color and update the text box for HSV and RGB
	for k,v in pairs(tColors) do
		self:SetColorValue(k, v * self.tValues[k].Max)
	end
	
	self.tValues[k_Hex]:SetText(self.ColorUtils:RgbToString(self:GetColor(k_R), self:GetColor(k_G), self:GetColor(k_B)))

	-- If this is coming from a text box update, then also update the Hue and Overlay
	if bUpdateCoords == true then
		self:SetProperty(k_Hue, 	self:GetColor(k_H) * self:GetWidth(k_Hue), 	   self:GetColor(k_H) * self:GetHeight(k_Hue), true)
		self:SetProperty(k_Overlay, self:GetColor(k_S) * self:GetWidth(k_Overlay), (1 - self:GetColor(k_V)) * self:GetHeight(k_Overlay), true)
	end
end	

-----------------------------------------------------------------------------------------------
-- Callback for text box updates for HSV/RGB.
-----------------------------------------------------------------------------------------------
function IronColorPicker:OnValueTextEdited(wndHandler, wndControl, strText)
-----------------------------------------------------------------------------------------------
	-- Handle 0 length strings
	if string.len(strText) == 0 then
		strText = "0"
	end

	-- Check to make sure every character is a digit
	local str = ""
	for c in strText:gmatch"." do
		if string.match(c, "%d") == nil then
			break
		end
		-- append digits onto the end of the string
		str = str .. c
	end

	-- Update all of the values
	self:SetColorValue(wndHandler:GetName(), tonumber(str, 10))
	self:UpdateValues(wndHandler:GetName(), true)
end

-----------------------------------------------------------------------------------------------
-- Callback for increment/decrement buttons on HSV/RGB.
-----------------------------------------------------------------------------------------------
function IronColorPicker:OnValueIncOrDec(wndHandler)
-----------------------------------------------------------------------------------------------
	local name = wndHandler:GetName()
	local value = (name == "Inc") and 1 or -1
	wndHandler:GetData()(value)
end	

-----------------------------------------------------------------------------------------------
-- Mouse Handling
-- OnMouseMove
-- OnMouseDown
-- OnMouseUp
-----------------------------------------------------------------------------------------------

-----------------------------------------------------------------------------------------------
-- Handles movement inside the window, and sets the X and Y positions for each window.
-----------------------------------------------------------------------------------------------
function IronColorPicker:OnMouseMove(wndHandler, wndControl, nLastRelativeMouseX, nLastRelativeMouseY)
-----------------------------------------------------------------------------------------------
	local name = wndHandler:GetName()
	if self:GetMouse(name) or name == k_Main then
		self:SetProperty(name, nLastRelativeMouseX, nLastRelativeMouseY, false)
	end
end

-----------------------------------------------------------------------------------------------
-- Handles Mouse Left Click down, and sets the X and Y position as well as the 
-- mouse down for the window.
-----------------------------------------------------------------------------------------------
function IronColorPicker:OnMouseDown( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
-----------------------------------------------------------------------------------------------
	local bLeftMouse = (eMouseButton == k_LeftMouse)
	local name       = wndHandler:GetName()

	if bLeftMouse == true then
		self:SetMouse(name, true)
		self:SetProperty(name, nLastRelativeMouseX, nLastRelativeMouseY, false)
	end
end

-----------------------------------------------------------------------------------------------
-- Handles Mouse Left Click up, and resets the mouse down for all windows.
-----------------------------------------------------------------------------------------------
function IronColorPicker:OnMouseUp( wndHandler, wndControl, eMouseButton, nLastRelativeMouseX, nLastRelativeMouseY, bDoubleClick, bStopPropagation )
-----------------------------------------------------------------------------------------------
	local bLeftMouse = (eMouseButton == k_LeftMouse)
	local name       = wndHandler:GetName()
	
	if bLeftMouse == true then
		self:SetMouse(k_Hue,     false)
		self:SetMouse(k_Overlay, false)
		self:SetMouse(k_Main,    false)
	end
end

-----------------------------------------------------------------------------------------------
-- UI Initialization
-- LoadCoord
-- LoadWidget
-- LoadValues
-- LoadHex
-----------------------------------------------------------------------------------------------

-----------------------------------------------------------------------------------------------
-- Loads a window that needs coordinates, and keeps track of the initial values.
-- All of the windows are stored in a class table called tCoords.
-----------------------------------------------------------------------------------------------
function IronColorPicker:LoadCoord(strName, tParams)
-----------------------------------------------------------------------------------------------
	local widget 				 = self.wndMain:FindChild(strName)
	assert(widget)

	self.tCoords[strName]        = {}
	self.tCoords[strName].Widget = widget
	self.tCoords[strName].Height = widget:GetHeight()
	self.tCoords[strName].Width  = widget:GetWidth()
	self.tCoords[strName].X      = 0
	self.tCoords[strName].Y      = 0
	self.tCoords[strName].Mouse  = false

	if tParams == nil then return end

	if tParams.strDefaultSprite then
		self.tCoords[strName].Widget:SetSprite(tParams.strDefaultSprite)
	end

	if tParams.nDefaultBGColor then
		self.tCoords[strName].Widget:SetBGColor(tParams.nDefaultBGColor)
	end

	if tParams.nDefaultRotation then
		self.tCoords[strName].Widget:SetRotation(tParams.nDefaultRotation)
	end
end

-----------------------------------------------------------------------------------------------
function IronColorPicker:LoadWidget(parentForm, name, tSub)
-----------------------------------------------------------------------------------------------
	local tWidget = {}
	local form = Apollo.LoadForm(self.xmlDoc, name, parentForm, self)
	tWidget.Widget = form
	for i,v in ipairs(tSub) do
		tWidget[v] = form:FindChild(v)
	end
	return tWidget
end

-----------------------------------------------------------------------------------------------
-- Initializer for RBG and HSV color boxes
-----------------------------------------------------------------------------------------------
function IronColorPicker:LoadValues(strName, tVals)
-----------------------------------------------------------------------------------------------
	local holder = self.wndMain:FindChild(strName)

	for i,v in ipairs(tVals) do
		local tWidget = self:LoadWidget(holder, "ValueBox", {"Value", "Description", "Inc", "Dec"})
		tWidget.Value:SetName(v.name)
		tWidget.Value:SetMaxTextLength(3)
		tWidget.Description:SetText(v.name)
		self.tValues[v.name] = {}
		self.tValues[v.name].Widget = tWidget.Value
		self.tValues[v.name].Max    = v.max
		self.tValues[v.name].Min    = v.min
		self.tValues[v.name].Value  = v.value/v.max

		local callback = function(val)
			self:SetColorValue(v.name, self:GetColorValue(v.name) + val)
			self:UpdateValues(v.name, true)
		end

		tWidget.Inc:SetData(callback)
		tWidget.Dec:SetData(callback)
	end

	holder:ArrangeChildrenVert(Window.CodeEnumArrangeOrigin.Middle)
end

-----------------------------------------------------------------------------------------------
function IronColorPicker:LoadHex(strName, strValue)
-----------------------------------------------------------------------------------------------
	local hex = self.wndMain:FindChild(strName)
	local widget = hex:FindChild("Value")
	widget:SetText(strValue)
	self.tValues[strName] = widget
end	

-----------------------------------------------------------------------------------------------
-- Accessors for properties of UI elements
-----------------------------------------------------------------------------------------------
function IronColorPicker:GetWidth(strName)  	   return self.tCoords[strName].Width  		end	-- READ_ONLY
function IronColorPicker:GetHeight(strName) 	   return self.tCoords[strName].Height 		end -- READ_ONLY
function IronColorPicker:GetX(strName, nValue) 	   return self.tCoords[strName].X   		end		
function IronColorPicker:GetY(strName, nValue) 	   return self.tCoords[strName].Y   		end
function IronColorPicker:GetMouse(strName) 		   return self.tCoords[strName].Mouse 		end
function IronColorPicker:GetWidget(strName)        return self.tCoords[strName].Widget      end
function IronColorPicker:GetColor(strName)         return self.tValues[strName].Value end
function IronColorPicker:GetColorValue(strName)    return self:GetColor(strName) * self.tValues[strName].Max end
function IronColorPicker:SetX(strName, nValue)     self.tCoords[strName].X = nValue 		end		
function IronColorPicker:SetY(strName, nValue)     self.tCoords[strName].Y = nValue			end
function IronColorPicker:SetMouse(strName, bValue) self.tCoords[strName].Mouse = bValue     end
function IronColorPicker:SetColor(strName, nValue) self.tValues[strName].Value = nValue     end
-----------------------------------------------------------------------------------------------

-----------------------------------------------------------------------------------------------
-- Generate a color table for HSV or RBG
-----------------------------------------------------------------------------------------------
function IronColorPicker:GetHSVTable()
-----------------------------------------------------------------------------------------------
	return { H = self:GetColor(k_H), S = self:GetColor(k_S), V = self:GetColor(k_V) }
end	

-----------------------------------------------------------------------------------------------
function IronColorPicker:GetRGBTable()
-----------------------------------------------------------------------------------------------
	return {R = self:GetColor(k_R), G = self:GetColor(k_G),  B = self:GetColor(k_B) }
end

-----------------------------------------------------------------------------------------------
-- IronColorPicker Instance
-----------------------------------------------------------------------------------------------
local IronColorPickerInst = IronColorPicker:new()
IronColorPickerInst:Init()
